package cd.casic.ci.process.api.engine.atom;

import cd.casic.ci.common.pipeline.enums.BuildStatus;
import cd.casic.ci.common.pipeline.enums.ErrorCode;
import cd.casic.ci.common.pipeline.pojo.ErrorType;
import cd.casic.ci.common.pipeline.utils.EnvUtils;
import cd.casic.ci.event.enums.ActionType;
import cd.casic.ci.process.api.engine.common.Timeout;
import cd.casic.ci.process.api.engine.common.VMUtils;
import cd.casic.ci.process.api.engine.pojo.PipelineBuildTask;
import cn.hutool.core.date.DateUtil;
import lombok.Getter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import static cd.casic.ci.common.pipeline.pojo.element.ElementAdditionalOptions.RunCondition;

/**
 * @author by mianbin
 * @Classname IAtomTask
 * @Description TODO
 * @Date 2025/3/31 16:01
 */
public interface IAtomTask<T> {

    Logger logger = LoggerFactory.getLogger(IAtomTask.class);
    /**
     * 开始执行原子逻辑,一般原子的业务逻辑都在这里处理.
     * 请注意: 如果有需要线程大量的循环sleep等待其他处理结果来判断是否成功的逻辑，请拆开到 tryFinish 去做判断
     * 并且在 tryFinish 函数确认返回false,则引擎后续会继续调度 tryFinish 函数来处理.
     * 该设计用来处理那些需要大量循环查询原子任务是否结束的，拆分为单独的 tryFinish 进行尝试查询状态以终止，减少线程被独占做无用的sleep
     * 应用场景：
     * 1.同步原子，需要轮循等待接口返回数据以决定原子是否完成整个业务, eg: CodeCC/apk加固/子流水线调用
     *
     * 例外： 当前原子执行失败则不会等待，直接标识为当前原子执行结束 status.isFinish()
     *
     * @param task 执行任务
     * @param param 参数
     * @param runVariables 当前流水线运行中产生的变量
     * @return BuildStatus 返回标志位是否要等待其他任务结束才结束
     */
    AtomResponse execute(PipelineBuildTask task, T param, Map<String, String> runVariables);

    /**
     * 读取参数Element
     */
    T getParamElement(PipelineBuildTask task);

    default AtomResponse execute(PipelineBuildTask task, Map<String, String> runVariables) {
        return execute(task, getParamElement(task), runVariables);
    }

    /**
     * 是否结束当前原子结束，如果不是，则后续引擎会不断地去调用该方法来判断原子是否结束了。
     * 子类通过复写该方法来实现，目前默认用查任务状态的方式保证大部分原子都是这类的操作
     *
     *
     * @param task 执行任务
     * @param runVariables 运行时变量
     * @param actionType 事件动作
     * @return BuildStatus
     *      返回标志位是否要等待其他任务结束才结束，如果返回 status.isFinish()
     *          true: 可以结束当前原子
     *          false: 需要等待其在他任务执行完。后续会不断的去调用该函数去查直到false，或者超时
     *          例外: 当前原子执行失败则不会等待，直接标识为当前原子执行结束
     */
    default AtomResponse tryFinish(PipelineBuildTask task, Map<String, String> runVariables, ActionType actionType) {
        T param = getParamElement(task);
        AtomResponse atomResponse = tryFinishImpl(task, param, runVariables, actionType);
        // 未结束？检查是否超时
        if (!atomResponse.getBuildStatus().isFinish()) {
            long startTime = task.getStartTime() != null ? DateUtil.toInstant(task.getStartTime()).toEpochMilli() : 0L;
            Long timeout = task.getAdditionalOptions() != null ? task.getAdditionalOptions().getTimeout() : null;
            if (timeout == null && VMUtils.isVMTask(task.getTaskId())) {
                // 如果timeout为空且task为开关机插件任务，则给timeout赋默认值
                timeout = Timeout.DEFAULT_PREPARE_MINUTES;
            }
            long timeoutMills = Timeout.transMinuteTimeoutToMills(timeout);
            RunCondition runCondition = task.getAdditionalOptions() != null ? task.getAdditionalOptions().getRunCondition() : null;
            if (timeoutMills > 0 && System.currentTimeMillis() - startTime >= timeoutMills) {
                logger.info(
                        "[{}]|TIME_OUT|startTime={}|timeoutMills={}|current={}",
                        task.getBuildId(), startTime, timeoutMills, System.currentTimeMillis()
                );
                atomResponse = new AtomResponse(
                        BuildStatus.EXEC_TIMEOUT,
                        null,
                        ErrorType.USER,
                        ErrorCode.USER_TASK_OUTTIME_LIMIT,
                        "Task time out " + TimeUnit.MILLISECONDS.toMinutes(timeoutMills) + " minutes"
                );
            } else if (actionType.isTerminate()) { // 强制终止的设置为失败
                logger.info("[{}]|FORCE_TERMINATE|job={}|task={}", task.getBuildId(), task.getContainerId(), task.getTaskId());
                atomResponse = defaultFailAtomResponse;
            } else if (actionType == ActionType.END && runCondition != RunCondition.PRE_TASK_FAILED_EVEN_CANCEL) {
                logger.info("[{}]|CANCEL|job={}|task={}", task.getBuildId(), task.getContainerId(), task.getTaskId());
                atomResponse = new AtomResponse(BuildStatus.CANCELED, null, null, null, null);
            }
        }
        return atomResponse;
    }

    private AtomResponse tryFinishImpl(
            PipelineBuildTask task,
            T param,
            Map<String, String> runVariables,
            ActionType actionType
    ) {
        AtomResponse atomResponse = tryFinish(task, param, runVariables, actionType.isTerminate());
        RunCondition runCondition = task.getAdditionalOptions() != null ? task.getAdditionalOptions().getRunCondition() : null;
        return !atomResponse.getBuildStatus().isFinish()
                ? actionType.isTerminate()
                ? new AtomResponse(
                BuildStatus.FAILED,
                null,
                ErrorType.PLUGIN,
                ErrorCode.PLUGIN_DEFAULT_ERROR,
                "Force Terminate!"
        )
                : actionType == ActionType.END && runCondition != RunCondition.PRE_TASK_FAILED_EVEN_CANCEL
                ? new AtomResponse(BuildStatus.CANCELED, null, null, null, null)
                : atomResponse
                : atomResponse;
    }

    default AtomResponse tryFinish(
            PipelineBuildTask task,
            T param,
            Map<String, String> runVariables,
            boolean force
    ) {
        return force
                ? task.getStatus().isFinish()
                ? new AtomResponse(
                task.getStatus(),
                null,
                task.getErrorType(),
                task.getErrorCode(),
                task.getErrorMsg()
        )
                : defaultFailAtomResponse
                : new AtomResponse(
                task.getStatus(),
                null,
                task.getErrorType(),
                task.getErrorCode(),
                task.getErrorMsg()
        );
    }

    default String parseVariable(String value, Map<String, String> runVariables) {
        if (value == null || value.isEmpty()) {
            return "";
        }
        return EnvUtils.parseEnv(value, runVariables);
    }

    /**
     * 原子执行结果
     * @version 1.0
     */
    @Getter
    class AtomResponse {
        private final BuildStatus buildStatus;
        private final Map<String, Object> outputVars;
        private ErrorType errorType;
        private Integer errorCode;
        private String errorMsg;

        public AtomResponse(BuildStatus buildStatus, Map<String, Object> outputVars, ErrorType errorType, Integer errorCode, String errorMsg) {
            this.buildStatus = buildStatus;
            this.outputVars = outputVars;
            this.errorType = errorType;
            this.errorCode = errorCode;
            this.errorMsg = errorMsg;
        }
    }

    AtomResponse defaultSuccessAtomResponse = new AtomResponse(BuildStatus.SUCCEED, null, null, null, null);

    AtomResponse defaultFailAtomResponse = new AtomResponse(
            BuildStatus.FAILED,
            null,
            ErrorType.USER,
            ErrorCode.PLUGIN_DEFAULT_ERROR,
            "not definded error"
    );
}
